from rougail.annotator.variable import Walk
from rougail.error import DictConsistencyError
from rougail.utils import normalize_family
from risotto.utils import _
from warnings import warn
from typing import List, Tuple


class Annotator(Walk):
    level = 9
    def __init__(self,
                 objectspace: 'RougailObjSpace',
                 *args):
        self.providers = {}
        self.suppliers = {}
        self.globals = {}
        self.provider_links = {}
        self.providers_zone = {}
        self.provider_maps = {}
        self.objectspace = objectspace
        #
        self.get_providers_suppliers()
        self.get_provider_links()
        self.get_provider_maps()
        self.convert_providers()
        self.convert_globals()

    def get_providers_suppliers(self) -> None:
        """parse all variable and get provider and supplier informations
        """
        for variable in self.get_variables():
            for type_ in 'provider', 'supplier':
                if hasattr(variable, type_):
                    provider_name = getattr(variable, type_)
                    # add information with provider/supplier name
                    if not hasattr(variable, 'information'):
                        variable.information = self.objectspace.information(variable.xmlfiles)
                    setattr(variable.information, type_, provider_name)

                    # construct self.globals, self.suppliers and self.providers dictionnaries
                    provider_prefix, provider_suffix = self._cut_out_provider_name(provider_name)
                    dns = self.objectspace.space.variables[variable.path_prefix].doc
                    if provider_prefix == 'global':
                        if type_ == 'supplier':
                            raise DictConsistencyError(f'{type_} {provider_name} in {dns} not allowed', 0, variable.xmlfiles)
                        obj = self.globals
                    elif type_ == 'supplier':
                        obj = self.suppliers
                    else:
                        obj = self.providers
                    sub_obj = obj.setdefault(provider_prefix, {}).setdefault(dns, {})
                    if provider_suffix in sub_obj:
                        raise DictConsistencyError(f'multiple {type_} {provider_name} in {dns}', 0, sub_obj[provider_suffix].xmlfiles + variable.xmlfiles)
                    sub_obj[provider_suffix] = variable

    def _cut_out_provider_name(self,
                               provider_name: str,
                               ) -> Tuple[str, str]:
        """get provider_name and return provider_prefix and provider_suffix
        """
        if ':' in provider_name:
            provider_prefix, provider_suffix = provider_name.split(':', 1)
        else:
            provider_prefix = provider_name
            provider_suffix = None
        return provider_prefix, provider_suffix

    def get_provider_links(self):
        """Search link between providers
           'ProviderPrefix': {'provider_dns': ['supplier_dns_1',
                                               'supplier_dns_2',
                                               'supplier_dns_3']}
        """
        for provider_prefix, providers_dns in self.providers.items():
            self.provider_links[provider_prefix] = {}
            for provider_dns, providers_suffix in providers_dns.items():
                if None not in providers_suffix:
                    # it's a reverse provider!
                    continue
                if provider_prefix != 'Host':
                    provider_zone = self.objectspace.rougailconfig['risotto_globals'][provider_dns]['global:provider_zone']
                if provider_prefix not in self.suppliers:
                    continue
                for supplier_dns, suppliers_suffix in self.suppliers[provider_prefix].items():
                    if provider_dns == supplier_dns:
                        continue
                    if provider_prefix == 'Host':
                        provider_zone = self.objectspace.rougailconfig['risotto_globals'][supplier_dns]['global:zones_name'][0]
                    if provider_zone not in self.objectspace.rougailconfig['risotto_globals'][supplier_dns]['global:zones_name']:
                        continue
                    self.provider_links[provider_prefix].setdefault(provider_dns, []).append(supplier_dns)

    def get_provider_maps(self):
        """relation between provider_prefix and provider_suffix
           'provider_prefix': {'normal': {None, 'provider_suffix_1', 'provider_suffix_2'},
                               'reverse': {'provider_suffix_3', 'provider_suffix_4'}}
        """
        for provider_prefix, providers_dns in self.provider_links.items():
            self.provider_maps[provider_prefix] = {'normal': set(), 'reverse': set()}
            for provider_dns, suppliers_dns in providers_dns.items():
                for supplier_dns in suppliers_dns:
                    if supplier_dns in self.providers[provider_prefix] and None in self.providers[provider_prefix][supplier_dns]:
                        #exists?
                        continue
                    # get prefixes
                    prefixes = set(self.providers[provider_prefix][provider_dns]) & set(self.suppliers[provider_prefix][supplier_dns])
                    self.provider_maps[provider_prefix]['normal'] |= prefixes
                    # get suffixes
                    if supplier_dns not in self.providers[provider_prefix]:
                        continue
                    suffixes = set(self.providers[provider_prefix][supplier_dns]) & set(self.suppliers[provider_prefix][provider_dns])
                    self.provider_maps[provider_prefix]['reverse'] |= suffixes

    def convert_providers(self) -> None:
        """Convert providers informations to default values or fills
        """
        for provider_prefix, providers_dns in self.provider_links.items():
            for provider_dns, suppliers_dns in providers_dns.items():
                for provider_suffix in self.provider_maps[provider_prefix]['normal']:
                    self._convert_providers_normal(provider_prefix,
                                                   provider_suffix,
                                                   provider_dns,
                                                   suppliers_dns,
                                                   )
                for provider_suffix in self.provider_maps[provider_prefix]['reverse']:
                    self._convert_providers_reverse(provider_prefix,
                                                    provider_suffix,
                                                    provider_dns,
                                                    suppliers_dns,
                                                    )

    def _convert_providers_normal(self,
                                  provider_prefix: str,
                                  provider_suffix: str,
                                  provider_dns: str,
                                  suppliers_dns: dict,
                                  ) -> None:
        if provider_prefix != 'Host':
            provider_zone = self.objectspace.rougailconfig['risotto_globals'][provider_dns]['global:provider_zone']
            provider_option_dns = self._get_dns_from_provider_zone(provider_dns, provider_zone)
        variable = self.providers[provider_prefix][provider_dns][provider_suffix]
        if hasattr(variable, 'value'):
            raise DictConsistencyError(f'variable {variable.path} has a provider and a value', 0, variable.xmlfiles)
        suppliers_var = {}
        for supplier_dns in suppliers_dns:
            if provider_suffix not in self.suppliers[provider_prefix][supplier_dns]:
                continue
            if provider_prefix == 'Host':
                provider_zone = self.objectspace.rougailconfig['risotto_globals'][supplier_dns]['global:zones_name'][0]
                provider_option_dns = self._get_dns_from_provider_zone(provider_dns, provider_zone)
            supplier_variable = self.suppliers[provider_prefix][supplier_dns][provider_suffix]
            supplier_option_dns = self._get_dns_from_provider_zone(supplier_dns, provider_zone)
            suppliers_var[supplier_option_dns] = supplier_variable
        if provider_suffix:
            AddFill(self.objectspace,
                    provider_prefix,
                    provider_suffix,
                    provider_option_dns,
                    variable,
                    suppliers_var,
                    False,
                    supplier_variable.path_prefix,
                    )
        else:
            self._set_provider_supplier(provider_option_dns,
                                        variable,
                                        suppliers_var,
                                        )

    def _convert_providers_reverse(self,
                                   provider_prefix: str,
                                   provider_suffix: str,
                                   provider_dns: str,
                                   suppliers_dns: dict,
                                   ) -> None:
        if provider_prefix != 'Host':
            provider_zone = self.objectspace.rougailconfig['risotto_globals'][provider_dns]['global:provider_zone']
            provider_option_dns = self._get_dns_from_provider_zone(provider_dns, provider_zone)
        variable = self.suppliers[provider_prefix][provider_dns][provider_suffix]
        if hasattr(variable, 'value'):
            raise DictConsistencyError(f'variable {variable.path} has a provider and a value', 0, variable.xmlfiles)
        for supplier_dns in suppliers_dns:
            if provider_prefix == 'Host':
                provider_zone = self.objectspace.rougailconfig['risotto_globals'][supplier_dns]['global:zones_name'][0]
                provider_option_dns = self._get_dns_from_provider_zone(provider_dns, provider_zone)
            supplier_variable = self.providers[provider_prefix][supplier_dns][provider_suffix]
            supplier_option_dns = self._get_dns_from_provider_zone(supplier_dns, provider_zone)
            AddFill(self.objectspace,
                    provider_prefix,
                    provider_suffix,
                    supplier_option_dns,
                    supplier_variable,
                    {provider_option_dns: variable},
                    True,
                    supplier_variable.path_prefix,
                    )

    def _get_dns_from_provider_zone(self,
                                    dns,
                                    zone,
                                    ) -> str:
        risotto_global = self.objectspace.rougailconfig['risotto_globals'][dns]
        index = risotto_global['global:zones_name'].index(zone)
        return risotto_global['global:server_names'][index]

    def _set_provider_supplier(self,
                               provider_dns: str,
                               variable,
                               suppliers,
                               ) -> None:
        # suffix is None so not a client and there only one provider
        # the value of this variable is the list of suppliers DNS name
        if not variable.multi:
            raise DictConsistencyError(f'"{variable.name}" is a provider and must be a multi', 0, variable.xmlfiles)
        variable.default = list(suppliers)
        # suffix is None so the supplier values are provider DNS
        for sub_variable in suppliers.values():
            #FIXME
            #if hasattr(option, 'value'):
            #    raise DictConsistencyError(f'"{option.name}" is a supplier and cannot have value', 0, option.xmlfiles)
            if sub_variable.multi:
                raise DictConsistencyError(f'"{sub_variable.name}" is a supplier and mustnot be a multi', 0, sub_variable.xmlfiles)
            sub_variable.default = provider_dns

    def convert_globals(self):
        """Convert providers global informations to default values or fills
        """
        provider_prefix = 'global'
        for provider_dns, providers_suffix in self.globals[provider_prefix].items():
            for provider_suffix, variable in providers_suffix.items():
                provider_name = f'{provider_prefix}:{provider_suffix}'
                if provider_name not in self.objectspace.rougailconfig['risotto_globals'][provider_dns]:
                    raise DictConsistencyError(f'cannot find {provider_name} for variable {variable.path}, should be in {list(self.objectspace.rougailconfig["risotto_globals"][provider_dns])}', 0, variable.xmlfiles)
                provider_values = self.objectspace.rougailconfig['risotto_globals'][provider_dns][provider_name]
                if isinstance(provider_values, list) and self.objectspace.paths.is_dynamic(variable):
                    if variable.multi:
                        raise DictConsistencyError(f'variable {variable.path} has provider {provider_name} and is in dynamic family so must not be a multi', 0, variable.xmlfiles)
                    self._set_global_dynamic_option(variable, provider_values)
                else:
                    if isinstance(provider_values, list) and not variable.multi:
                        raise DictConsistencyError(f'variable {variable.path} has provider {provider_name} which is a multi', 0, variable.xmlfiles)
                    if not isinstance(provider_values, list):
                        if variable.multi:
                            raise DictConsistencyError(f'variable {variable.path} has provider {provider_name} which is not a multi', 0, variable.xmlfiles)
                        provider_values = [provider_values]

                    variable.value = []
                    for provider_value in provider_values:
                        value = self.objectspace.value(variable.xmlfiles)
                        value.name = provider_value
                        if isinstance(provider_value, bool):
                            value.type = 'boolean'
                        variable.value.append(value)

    def _set_global_dynamic_option(self,
                                   variable: 'self.objectspace.Variable',
                                   values: List[str],
                                   ):
        fill = self.objectspace.fill(variable.xmlfiles)
        new_target = self.objectspace.target(variable.xmlfiles)
        new_target.name = variable.name
        fill.target = [new_target]
        fill.namespace = variable.namespace
        fill.index = 0
        fill.name = 'risotto_providers_global'
        param1 = self.objectspace.param(variable.xmlfiles)
        param1.text = values
        param2 = self.objectspace.param(variable.xmlfiles)
        param2.type = 'suffix'
        fill.param = [param1, param2]
        if not hasattr(self.objectspace.space.variables[variable.path_prefix].constraints, 'fill'):
            self.objectspace.space.variables[variable.path_prefix].constraints.fill = []
        self.objectspace.space.variables[variable.path_prefix].constraints.fill.append(fill)


class AddFill:
    """Add fill for variable
    """
    def __init__(self,
                 objectspace,
                 provider_prefix,
                 provider_suffix,
                 provider_dns,
                 variable,
                 suppliers,
                 reverse,
                 path_prefix,
                 ) -> None:
        self.objectspace = objectspace
        self.provider_dns = provider_dns
        self.variable = variable
        self.path_prefix = path_prefix
        self.suppliers = suppliers
        self.create_fill()
        if reverse:
            self.param_reverse()
        else:
            if self.objectspace.paths.is_dynamic(self.variable):
                self.param_dynamic()
            elif self.variable.multi:
                self.param_multi()
            else:
                provider_name = f'{provider_prefix}:{provider_suffix}'
                raise DictConsistencyError(f'provider "{provider_name}" options must be in dynamic option or must be a multiple', 0, self.variable.xmlfiles)
            if self.objectspace.paths.is_follower(self.variable):
                self.param_follower()
        self.end()

    def create_fill(self) -> None:
        self.fill = self.objectspace.fill(self.variable.xmlfiles)
        new_target = self.objectspace.target(self.variable.xmlfiles)
        new_target.name = self.variable
        self.fill.target = [new_target]
        self.fill.namespace = self.variable.namespace
        self.fill.index = 0
        self.fill.param = []

    def param_reverse(self) -> None:
        self.fill.name = 'risotto_flatten_values_client'
        #
        if self.objectspace.paths.is_follower(self.variable):
            multi = self.variable.multi is True
        else:
            multi = self.variable.multi is not False
        param = self.objectspace.param(self.variable.xmlfiles)
        param.text = multi
        param.type = 'boolean'
        self.fill.param.append(param)
        for dns, variable in self.suppliers.items():
            param = self.objectspace.param(variable.xmlfiles)
            param.text = variable
            param.propertyerror = False
            param.type = 'variable'
            param.suffix = normalize_family(self.provider_dns)
            namespace = variable.namespace
            family_path = self.objectspace.paths.get_variable_family_path(param.text.path, 
                                                                          namespace,
                                                                          force_path_prefix=self.variable.path_prefix,
                                                                          )
            param.family = self.objectspace.paths.get_family(family_path,
                                                             namespace,
                                                             self.variable.path_prefix,
                                                             )
            self.fill.param.append(param)

    def param_dynamic(self) -> None:
        self.fill.name = 'risotto_dyn_values'
        #
        param = self.objectspace.param(self.variable.xmlfiles)
        param.type = 'suffix'
        self.fill.param.append(param)
        #
        param = self.objectspace.param(self.variable.xmlfiles)
        param.text = self.variable.unique != "False"
        param.type = 'boolean'
        self.fill.param.append(param)
        #
        if self.objectspace.paths.is_follower(self.variable):
            multi = self.variable.multi is True
        else:
            multi = self.variable.multi is not False
        param = self.objectspace.param(self.variable.xmlfiles)
        param.text = multi
        param.type = 'boolean'
        self.fill.param.append(param)
        for dns, variable in self.suppliers.items():
            #
            param = self.objectspace.param(variable.xmlfiles)
            param.text = variable
            param.name = normalize_family(dns)
            param.propertyerror = False
            param.type = 'variable'
            self.fill.param.append(param)

    def param_multi(self) -> None:
        self.fill.name = 'risotto_flatten_values'
        #
        if self.objectspace.paths.is_follower(self.variable):
            multi = self.variable.multi is True
        else:
            multi = self.variable.multi is not False
        param = self.objectspace.param(self.variable.xmlfiles)
        param.text = multi
        param.type = 'boolean'
        self.fill.param.append(param)
        for dns, variable in self.suppliers.items():
            param = self.objectspace.param(variable.xmlfiles)
            param.text = variable
            param.propertyerror = False
            param.type = 'variable'
            self.fill.param.append(param)

    def param_follower(self):
        param = self.objectspace.param(self.variable.xmlfiles)
        param.name = 'follower_index'
        param.type = 'index'
        self.fill.param.append(param)

    def end(self):
        if not hasattr(self.objectspace.space.variables[self.path_prefix], 'constraints'):
            self.objectspace.space.variables[self.path_prefix].constraints = self.objectspace.constraints(None)
        if not hasattr(self.objectspace.space.variables[self.path_prefix].constraints, 'fill'):
            self.objectspace.space.variables[self.path_prefix].constraints.fill = []
        self.objectspace.space.variables[self.path_prefix].constraints.fill.append(self.fill)
